@@ -1,3 +1,4 @@
+import io
 import logging
 import time
 import typing as t
@@ -8,8 +9,7 @@ from types import TracebackType
 import polars as pl
 from kedro.config import OmegaConfigLoader
 from kedro.framework.project import settings
-from sqlalchemy import create_engine, text, tuple_
-from sqlalchemy import text as sa_text
+from sqlalchemy import Column, MetaData, Table, create_engine, text, tuple_
 from sqlalchemy.dialects.postgresql import insert
 from sqlalchemy.engine import Engine
 from sqlalchemy.exc import SQLAlchemyError
@@ -307,7 +307,7 @@ class SQLAlchemyDatabase:
         data: pl.DataFrame,
         mode: t.Literal["RAISE", "IGNORE", "UPDATE"],
         *,
-        threshold_for_copy: int = 100_000_000_000,  # temporarily do not use copy
+        threshold_for_copy: int = 100000,  # temporarily do not use copy
         returning_id: bool = False,
     ) -> pl.DataFrame | None:
         """
@@ -352,7 +352,7 @@ class SQLAlchemyDatabase:
 
         Writes data first, then—if `returning_id=True`—fetches all primary keys
         for the provided unique-key combinations in one go.
-        Returns a DataFrame of unique-key columns plus "id" if requested,
+        Returns a DataFrame of unique-key columns plus "ID" if requested,
         otherwise returns None.
         """
         # Determine unique-key columns
@@ -380,7 +380,6 @@ class SQLAlchemyDatabase:
                 data,
                 chunk_size,
                 p_unique_key_columns,
-                table_name,
             )
 
         return None
@@ -403,10 +402,9 @@ class SQLAlchemyDatabase:
             tqdm(
                 row_iter,
                 total=total_rows,
-                desc=f"Writing to table: {table_name}",
+                desc=f"Bulk writing to table: {table_name}",
                 unit="rows",
                 dynamic_ncols=True,
-                mininterval=0.1,
             )
             if show_progress
             else row_iter
@@ -427,17 +425,13 @@ class SQLAlchemyDatabase:
             pass  # No ON CONFLICT clause needed
 
         with self.get_session() as session:
-            for i, offset in enumerate(row_iter):
+            for offset in row_iter:
                 chunk_df = data.slice(offset, chunk_size)
                 records = chunk_df.to_dicts()
                 session.execute(stmt, records)
                 # Manually update progress bar to show correct number of rows processed
                 if show_progress:
-                    # Calculate how many rows have been processed so far
-                    processed = min((i + 1) * chunk_size, total_rows)
-                    progress_bar.n = processed
-                    progress_bar.last_print_n = processed
-                    progress_bar.update(0)
+                    progress_bar.update(len(records))
 
     def _fetch_ids_chunks(
         self,
@@ -445,115 +439,133 @@ class SQLAlchemyDatabase:
         data: pl.DataFrame,
         chunk_size: int,
         p_unique_key_columns: list[str],
-        table_name: str,
-        use_temp_table: int = 1000,
-        profile: bool = False,
+        use_temp_table: int = 10,
     ) -> pl.DataFrame:
         """Helper to fetch IDs in chunks after writing data. Uses temp table for large chunks."""
-
-        def func(model_class, data, chunk_size, p_unique_key_columns, table_name, use_temp_table):
-            key_df = data.select(*p_unique_key_columns)
-            total_keys = key_df.height
-            total_chunks = (total_keys + chunk_size - 1) // chunk_size
-            row_iter = range(0, total_keys, chunk_size)
-            show_progress = total_chunks > 1
-            progress_bar = (
-                tqdm(
-                    row_iter,
-                    total=total_keys,
-                    desc=f"Fetching IDs from table: {table_name}",
-                    unit="rows",
-                    dynamic_ncols=True,
-                    mininterval=0.01,
-                )
-                if show_progress
-                else row_iter
+        key_df = data.select(*p_unique_key_columns)
+        total_keys = key_df.height
+        total_chunks = (total_keys + chunk_size - 1) // chunk_size
+        row_iter = range(0, total_keys, chunk_size)
+        show_progress = total_chunks > 1
+        progress_bar = (
+            tqdm(
+                row_iter,
+                total=total_keys,
+                desc=f"Fetching IDs from table: {model_class.__table__.name}",
+                unit="rows",
+                dynamic_ncols=True,
+            )
+            if show_progress
+            else row_iter
+        )
+        unique_key_attrs = [self.get_column_attr_by_db_name(model_class, col) for col in p_unique_key_columns]
+        use_temp = chunk_size > use_temp_table
+        if use_temp:
+            ids = self._fetch_ids_chunks_temp_table(
+                model_class,
+                key_df,
+                chunk_size,
+                p_unique_key_columns,
+                progress_bar,
             )
-            unique_key_attrs = [self.get_column_attr_by_db_name(model_class, col) for col in p_unique_key_columns]
-            ids = []
-            use_temp = chunk_size > use_temp_table
-            with self.get_session() as session:
-                # Precompute invariants for temp table approach
-                if use_temp:
-                    # Get column types from model (once)
-                    col_defs = []
-                    for col in p_unique_key_columns:
-                        col_obj = model_class.__table__.columns[col]
-                        col_type = col_obj.type.compile(dialect=session.bind.dialect)
-                        col_defs.append(f'"{col}" {col_type}')
-                    col_defs_sql = ", ".join(col_defs)
-                    col_names_sql = ", ".join([f'"{col}"' for col in p_unique_key_columns])
-                    param_names = [f":{col}" for col in p_unique_key_columns]
-                # Get the actual primary key column name from the model (once)
-                pk_col = next(iter(model_class.__table__.primary_key)).name
-                # Precompute join condition and select columns for temp table join
-                if use_temp:
-                    join_conds = " AND ".join(f'main."{col}" = temp."{col}"' for col in p_unique_key_columns)
-                    select_cols = ", ".join([f'main."{col}"' for col in p_unique_key_columns] + [f'main."{pk_col}"'])
-                for i, offset in enumerate(row_iter):
-                    chunk_keys = key_df.slice(offset, chunk_size)
-                    key_rows = chunk_keys.rows()
-                    if not key_rows:
-                        continue
-                    if use_temp:
-                        # Use temp table approach
-                        temp_table_name = f"temp_keys_{i}"
-                        # Drop if exists (safety)
-                    session.execute(text(f'DROP TABLE IF EXISTS "{temp_table_name}"'))
-                    # Use UNLOGGED for faster temp table creation
-                    session.execute(text(f'CREATE TEMP TABLE "{temp_table_name}" ({col_defs_sql}) ON COMMIT DROP'))
-                    # Bulk insert all rows at once
-                    insert_stmt = f'INSERT INTO "{temp_table_name}" ({col_names_sql}) VALUES ({", ".join(param_names)})'
-                    param_dicts = [dict(zip(p_unique_key_columns, row, strict=True)) for row in key_rows]
-                    if param_dicts:
-                        session.execute(sa_text(insert_stmt), param_dicts)
-                        # Join with main table using quoted column names
-                        join_sql = f'SELECT {select_cols} FROM "{table_name}" AS main JOIN "{temp_table_name}" AS temp ON {join_conds}'
-                        results = session.execute(text(join_sql)).fetchall()
-                        # Always use "id" as the output key for the PK
-                        for row in results:
-                            row_dict = dict(zip([*p_unique_key_columns, pk_col], row, strict=True))
-                            row_dict["id"] = row_dict.pop(pk_col)
-                            ids.append(row_dict)
-                        # Drop temp table (should be auto-dropped, but explicit for safety)
-                        session.execute(text(f'DROP TABLE IF EXISTS "{temp_table_name}"'))
-                        # ...handled above for temp table join...
-                    else:
-                        # Use IN clause approach
-                        key_tuples = [tuple(row) for row in key_rows]
-                        # Get the actual PK column name and attribute
-                        pk_col = next(iter(model_class.__table__.primary_key)).name
-                        pk_attr = getattr(model_class, pk_col)
-                        query = session.query(*unique_key_attrs, pk_attr)
-                        if len(p_unique_key_columns) == 1:
-                            col_attr = unique_key_attrs[0]
-                            query = query.filter(col_attr.in_(key_tuples))
-                        else:
-                            query = query.filter(tuple_(*unique_key_attrs).in_(key_tuples))
-                        results = query.all()
-                        for row in results:
-                            row_dict = dict(zip([*p_unique_key_columns, pk_col], row, strict=True))
-                            row_dict["id"] = row_dict.pop(pk_col)
-                            ids.append(row_dict)
-                    # Manually update progress bar to show correct number of rows processed
-                    if show_progress:
-                        processed = min((i + 1) * chunk_size, total_keys)
-                        progress_bar.n = processed
-                        progress_bar.last_print_n = processed
-                        progress_bar.update(0)
-            ids_df = pl.DataFrame(ids)
-            return ids_df
-
-        if profile:
-            from line_profiler import LineProfiler
-
-            profiler = LineProfiler()
-            profiled_func = profiler(func)
-            result = profiled_func(model_class, data, chunk_size, p_unique_key_columns, table_name, use_temp_table)
-            profiler.print_stats()
         else:
-            result = func(model_class, data, chunk_size, p_unique_key_columns, table_name, use_temp_table)
-        return result
+            ids = self._fetch_ids_chunks_in_clause(
+                model_class,
+                key_df,
+                chunk_size,
+                p_unique_key_columns,
+                unique_key_attrs,
+                progress_bar,
+            )
+        ids_df = pl.DataFrame(ids)
+        return ids_df
+
+    def _fetch_ids_chunks_in_clause(
+        self,
+        model_class: DeclarativeMeta,
+        key_df: pl.DataFrame,
+        chunk_size: int,
+        p_unique_key_columns: list[str],
+        unique_key_attrs: list[InstrumentedAttribute],
+        progress_bar: tqdm,
+    ) -> None:
+        """Fetch IDs using IN clause for each chunk."""
+        ids = []
+        with self.get_session() as session:
+            pk_col = next(iter(model_class.__table__.primary_key)).name
+            for offset in progress_bar:
+                chunk_keys = key_df.slice(offset, chunk_size)
+                key_rows = chunk_keys.rows()
+                if not key_rows:
+                    continue
+                key_tuples = [tuple(row) for row in key_rows]
+                pk_attr = self.get_column_attr_by_db_name(model_class, pk_col)
+                query = session.query(*unique_key_attrs, pk_attr)
+                if len(p_unique_key_columns) == 1:
+                    col_attr = unique_key_attrs[0]
+                    query = query.filter(col_attr.in_(key_tuples))
+                else:
+                    query = query.filter(tuple_(*unique_key_attrs).in_(key_tuples))
+                results = query.all()
+                for row in results:
+                    row_dict = dict(zip([*p_unique_key_columns, pk_col], row, strict=True))
+                    row_dict["ID"] = row_dict.pop(pk_col)
+                    ids.append(row_dict)
+                if hasattr(progress_bar, "update"):
+                    progress_bar.update(len(key_rows))
+        return ids
+
+    def _fetch_ids_chunks_temp_table(
+        self,
+        model_class: DeclarativeMeta,
+        key_df: pl.DataFrame,
+        chunk_size: int,
+        p_unique_key_columns: list[str],
+        progress_bar: tqdm,
+    ) -> None:
+        """Fetch IDs using a temporary table for each chunk."""
+        ids = []
+        with self.get_session() as session:
+            pk_col = next(iter(model_class.__table__.primary_key)).name
+            metadata = MetaData()
+            # Precompute column definitions, join conditions, and select columns
+            col_defs = []
+            for col in p_unique_key_columns:
+                col_obj = model_class.__table__.columns[col]
+                col_type = col_obj.type.compile(dialect=session.bind.dialect)
+                col_defs.append(f'"{col}" {col_type}')
+            col_defs_sql = ", ".join(col_defs)
+            join_conds = " AND ".join(f'main."{col}" = temp."{col}"' for col in p_unique_key_columns)
+            select_cols = ", ".join([f'main."{col}"' for col in p_unique_key_columns] + [f'main."{pk_col}"'])
+
+            for i, offset in enumerate(progress_bar):
+                chunk_keys = key_df.slice(offset, chunk_size)
+                key_rows = chunk_keys.rows()
+                if not key_rows:
+                    continue
+                temp_table_name = f"temp_keys_{i}"
+                session.execute(text(f'DROP TABLE IF EXISTS "{temp_table_name}"'))
+                session.execute(text(f'CREATE TEMP TABLE "{temp_table_name}" ({col_defs_sql}) ON COMMIT DROP'))
+                col_objs = []
+                for col in p_unique_key_columns:
+                    col_obj = model_class.__table__.columns[col]
+                    col_objs.append(Column(col, col_obj.type))
+                temp_table = Table(temp_table_name, metadata, *col_objs)
+                param_dicts = [dict(zip(p_unique_key_columns, row, strict=True)) for row in key_rows]
+                if param_dicts:
+                    session.execute(insert(temp_table), param_dicts)
+                    join_sql = f'SELECT {select_cols} FROM "{model_class.__table__.name}" AS main JOIN "{temp_table_name}" AS temp ON {join_conds}'  # noqa: S608
+                    results = session.execute(text(join_sql)).fetchall()
+                else:
+                    results = []
+                session.execute(text(f'DROP TABLE IF EXISTS "{temp_table_name}"'))
+                for row in results:
+                    row_dict = dict(zip([*p_unique_key_columns, pk_col], row, strict=True))
+                    row_dict["ID"] = row_dict.pop(pk_col)
+                    ids.append(row_dict)
+                if hasattr(progress_bar, "update"):
+                    progress_bar.update(len(key_rows))
+        return ids
 
     def handle_insertion_multi_line_copy(
         self,
@@ -561,10 +573,115 @@ class SQLAlchemyDatabase:
         data: pl.DataFrame,
         mode: t.Literal["RAISE", "IGNORE", "UPDATE"],
         *,
-        chunk_size: int = 1_000_000,
+        chunk_size: int = 100_000,
         returning_id: bool = False,
     ) -> pl.DataFrame | None:
-        raise NotImplementedError
+        """
+        Bulk-load via PostgreSQL COPY into a temp table, then upsert from that staging table.
+
+        - Splits the Polars DataFrame into batches of `chunk_size` rows.
+        - For each batch:
+            1. Creates a TEMP TABLE LIKE the target table (no indexes/constraints)
+            2. COPYs the batch into the temp table (CSV via STDIN)
+            3. INSERTs from temp into real table with ON CONFLICT configured by `mode`
+            4. If `returning_id=True`, SELECTs back all {pk_cols..., id} by joining real+temp
+        - Commits each batch independently, returning a combined DataFrame of IDs if requested.
+        """
+        # Primary-key columns used for conflict detection
+        pks = self.get_primary_unique_constraint(model_class)
+        table = model_class.__table__.name
+        cols = data.columns
+        col_list = ", ".join(f'"{c}"' for c in cols)
+
+        # Prepare batching
+        total_batches = (data.height + chunk_size - 1) // chunk_size
+        show_progress = total_batches > 1
+
+        logger.info(
+            "Starting COPY insert of %d rows in %d chunks of size %d",
+            data.height,
+            (data.height + chunk_size - 1) // chunk_size,
+            chunk_size,
+        )
+        if show_progress:
+            progress_bar = tqdm(
+                total=data.height,
+                desc=f"COPY writing to table: {table}",
+                unit="rows",
+                dynamic_ncols=True,
+            )
+
+        all_id_maps: list[pl.DataFrame] = []
+
+        # Use single connection for all batches
+        raw_conn = self.engine.raw_connection()
+        cursor = raw_conn.cursor()
+
+        try:
+            # Precompute ON CONFLICT clause based on mode
+            conflict_cols = ", ".join(f'"{c}"' for c in pks)
+            if mode == "RAISE":
+                on_conflict = ""
+            elif mode == "IGNORE":
+                on_conflict = f"ON CONFLICT ({conflict_cols}) DO NOTHING"
+            else:  # UPDATE
+                non_pk_cols = [c for c in cols if c not in pks]
+                if non_pk_cols:
+                    set_clause = ", ".join(f'"{c}" = EXCLUDED."{c}"' for c in non_pk_cols)
+                    on_conflict = f"ON CONFLICT ({conflict_cols}) DO UPDATE SET {set_clause}"
+                else:
+                    on_conflict = f"ON CONFLICT ({conflict_cols}) DO NOTHING"
+
+            # Process each batch
+            for batch_idx, offset in enumerate(range(0, data.height, chunk_size)):
+                batch = data.slice(offset, chunk_size)
+                if batch.height == 0:
+                    continue
+
+                temp_table = f"temp_{table}_{batch_idx}"
+
+                # 1) Create temp table (just structure, no constraints/indexes)
+                cursor.execute(f'CREATE TEMP TABLE "{temp_table}" (LIKE "{table}") ON COMMIT DROP')
+
+                # 2) COPY data into staging
+                buf = io.StringIO()
+                batch.write_csv(buf)
+                buf.seek(0)
+                cursor.copy_expert(f'COPY "{temp_table}" ({col_list}) FROM STDIN WITH CSV HEADER', buf)
+
+                # 3) INSERT FROM staging into real table
+                insert_sql = f'INSERT INTO "{table}" ({col_list}) SELECT {col_list} FROM "{temp_table}" ' + on_conflict  # noqa: S608
+                cursor.execute(insert_sql)
+
+                # 4) Optionally retrieve IDs by joining via COPY + Polars
+                if returning_id:
+                    buf_out = io.StringIO()
+                    # Build SELECT clause with explicit aliases
+                    select_list = ", ".join(['r."ID" AS "ID"', *[f't."{c}" AS "{c}"' for c in pks]])
+                    copy_sql = f'COPY (SELECT {select_list} FROM "{table}" AS r JOIN "{temp_table}" AS t USING ({conflict_cols})) TO STDOUT WITH CSV HEADER'  # noqa: S608
+                    cursor.copy_expert(copy_sql, buf_out)
+                    buf_out.seek(0)
+                    df_ids = pl.read_csv(buf_out)
+                    all_id_maps.append(df_ids)
+
+                # Drop temp table explicitly
+                cursor.execute(f'DROP TABLE IF EXISTS "{temp_table}"')
+
+                # Update progress bar with number of rows processed
+                if show_progress:
+                    progress_bar.update(batch.height)
+
+            # Final commit for all batches
+            raw_conn.commit()
+
+        finally:
+            cursor.close()
+            raw_conn.close()
+
+        if returning_id:
+            all_dfs = pl.concat(all_id_maps)
+            return all_dfs
+        return None
 
     @staticmethod
     def get_column_attr_by_db_name(model_class: DeclarativeMeta, db_col_name: str) -> InstrumentedAttribute:
